-- ==========================================================================
-- postbio.sql
-- Bio extensions for PostgreSQL
-- Author: Luis Carvalho <lexcarvalho at gmail dot com>
-- Please check copyright notice at the bottom of this file
-- $Id: postbio.sql.in,v 1.5 2009/09/19 16:24:05 carvalho Exp $
-- ==========================================================================

-- ==========================================================================
-- Int_interval
-- ==========================================================================

CREATE TYPE int_interval;

CREATE OR REPLACE FUNCTION iint_in (cstring)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_out (int_interval)
  RETURNS cstring
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_recv (internal)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_send (int_interval)
  RETURNS bytea
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE TYPE int_interval (
  internallength = 8,
  input = iint_in,
  output = iint_out,
  receive = iint_recv,
  send = iint_send,
  alignment = integer
);

CREATE OR REPLACE FUNCTION iint_low (int_interval)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_high (int_interval)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

-- Useful casts

CREATE OR REPLACE FUNCTION integer_to_iint (integer)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE CAST (integer AS int_interval)
  WITH FUNCTION integer_to_iint(integer) AS IMPLICIT;

CREATE OR REPLACE FUNCTION iint_textin (text)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE CAST (text AS int_interval)
  WITH FUNCTION iint_textin(text) AS IMPLICIT;

CREATE OR REPLACE FUNCTION iint_textout (int_interval)
  RETURNS text
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE CAST (int_interval AS text)
  WITH FUNCTION iint_textout(int_interval) AS IMPLICIT;

CREATE OR REPLACE FUNCTION intpair_to_iint (integer, integer)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

--CREATE OR REPLACE FUNCTION int_array_to_int (ia _int_interval)
--    RETURNS int_interval AS $$
--  BEGIN
--    RETURN ia[1];
--  END;
--  $$ LANGUAGE plpgsql IMMUTABLE STRICT;
--
--CREATE CAST (_int_interval AS int_interval)
--  WITH FUNCTION int_array_to_int(_int_interval) AS IMPLICIT;

-- B-tree and R-tree (indexing) strategies

CREATE OR REPLACE FUNCTION iint_left (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_overleft (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_overlaps (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_overright (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_right (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_same (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_contains (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_contained (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

-- B-tree support

CREATE OR REPLACE FUNCTION iint_compare (int_interval, int_interval)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_lt (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_le (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_eq (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_ge (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_gt (int_interval, int_interval)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

-- R-tree support

CREATE OR REPLACE FUNCTION g_iint_consistent (internal, int_interval, int4)
  RETURNS bool
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION g_iint_union (internal, internal)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION g_iint_compress (internal)
  RETURNS internal
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION g_iint_decompress (internal)
  RETURNS internal
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION g_iint_penalty (internal, internal, internal)
  RETURNS internal
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION g_iint_picksplit (internal, internal)
  RETURNS internal
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION g_iint_same (int_interval, int_interval, internal)
  RETURNS internal
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

-- Tools

CREATE OR REPLACE FUNCTION iint_distance_low (int_interval, int_interval)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_distance_high (int_interval, int_interval)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_distance_center (int_interval, int_interval)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_distance (int_interval, int_interval)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_join (int_interval, int_interval)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_overlap (int_interval, int_interval)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_size (int_interval)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_buffer (int_interval, integer)
  RETURNS int_interval
  AS '$libdir/postbio', 'iint_buffer1'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_buffer (int_interval, integer, integer)
  RETURNS int_interval
  AS '$libdir/postbio', 'iint_buffer2'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION iint_span (int_interval, int_interval)
  RETURNS int_interval
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;


-- Operators

-- B-tree

CREATE OPERATOR < (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_lt,
  commutator = '>',
  negator = '>=',
  restrict = scalarltsel,
  join = scalarltjoinsel
);

CREATE OPERATOR <= (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_le,
  commutator = '>=',
  negator = '>',
  restrict = scalarltsel,
  join = scalarltjoinsel
);

CREATE OPERATOR = (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_eq,
  commutator = '=',
  restrict = eqsel,
  join = eqjoinsel
);

CREATE OPERATOR >= (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_ge,
  commutator = '<=',
  negator = '<',
  restrict = scalargtsel,
  join = scalargtjoinsel
);

CREATE OPERATOR > (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_gt,
  commutator = '<',
  negator = '<=',
  restrict = scalargtsel,
  join = scalargtjoinsel
);


-- R-tree

CREATE OPERATOR << (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_left,
  commutator = '>>',
  negator = '&>',
  restrict = positionsel,
  join = positionjoinsel
);

CREATE OPERATOR &< (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_overleft,
  commutator = '&>',
  negator = '>>',
  restrict = positionsel,
  join = positionjoinsel
);

CREATE OPERATOR && (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_overlaps,
  commutator = '&&',
  restrict = areasel,
  join = areajoinsel
);

CREATE OPERATOR &> (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_overright,
  commutator = '&<',
  negator = '<<',
  restrict = positionsel,
  join = positionjoinsel
);

CREATE OPERATOR >> (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_right,
  commutator = '<<',
  negator = '&<',
  restrict = positionsel,
  join = positionjoinsel
);

CREATE OPERATOR ~= (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_same,
  commutator = '~=',
  restrict = areasel,
  join = areajoinsel
);

CREATE OPERATOR ~ (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_contains,
  commutator = '@',
  restrict = contsel,
  join = contjoinsel
);

CREATE OPERATOR @ (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_contained,
  commutator = '~',
  restrict = contsel,
  join = contjoinsel
);

-- Tools

CREATE OPERATOR <|-| (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_distance_low
);

CREATE OPERATOR |-|> (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_distance_high
);

CREATE OPERATOR @-@ (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_distance_center
);

CREATE OPERATOR |-| (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_distance
);

CREATE OPERATOR + (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_join,
  commutator = '+'
);

CREATE OPERATOR * (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_overlap,
  commutator = '*'
);

CREATE OPERATOR ^ (
  leftarg = int_interval,
  rightarg = int_interval,
  procedure = iint_span,
  commutator = '^'
);

CREATE OPERATOR # (
  rightarg = int_interval,
  procedure = iint_size
);

CREATE OPERATOR #< (
  rightarg = int_interval,
  procedure = iint_low
);

CREATE OPERATOR #> (
  rightarg = int_interval,
  procedure = iint_high
);

CREATE OPERATOR <-> (
  leftarg = int_interval,
  rightarg = integer,
  procedure = iint_buffer
);


-- Indexes

CREATE OPERATOR CLASS btree_iint_interval_ops
    DEFAULT FOR TYPE int_interval USING btree AS
  OPERATOR 1 < ,
  OPERATOR 2 <= ,
  OPERATOR 3 = ,
  OPERATOR 4 >= ,
  OPERATOR 5 > ,
  FUNCTION 1 iint_compare (int_interval, int_interval);

CREATE OPERATOR CLASS g_iint_interval_ops
    DEFAULT FOR TYPE int_interval USING gist AS
  OPERATOR 1 << ,
  OPERATOR 2 &< ,
  OPERATOR 3 && ,
  OPERATOR 4 &> ,
  OPERATOR 5 >> ,
  OPERATOR 6 ~= ,
  OPERATOR 7 ~ ,
  OPERATOR 8 @ ,
  FUNCTION 1 g_iint_consistent (internal, int_interval, int4),
  FUNCTION 2 g_iint_union (internal, internal),
  FUNCTION 3 g_iint_compress (internal),
  FUNCTION 4 g_iint_decompress (internal),
  FUNCTION 5 g_iint_penalty (internal, internal, internal),
  FUNCTION 6 g_iint_picksplit (internal, internal),
  FUNCTION 7 g_iint_same (int_interval, int_interval, internal);

-- Aggregates

CREATE AGGREGATE iint_union (
  sfunc = iint_join,
  basetype = int_interval,
  stype = int_interval
);

CREATE AGGREGATE iint_intersection (
  sfunc = iint_overlap,
  basetype = int_interval,
  stype = int_interval
);

CREATE AGGREGATE iint_cover (
  sfunc = iint_span,
  basetype = int_interval,
  stype = int_interval
);

-- PL/pgSQL functions

--CREATE FUNCTION iint_merge(t text, f text)
--    RETURNS SETOF int_interval AS $$
--DECLARE
--  r int_interval;
--  u int_interval;
--  c int_interval;
--BEGIN
--  -- prepare query on table t, field f
--  CREATE TEMPORARY TABLE temp (interv int_interval, visited boolean);
--  EXECUTE 'INSERT INTO temp SELECT ' || quote_ident(f)
--    || ', FALSE FROM ' || quote_ident(t);
--  CREATE INDEX temp_idx ON temp USING gist (interv);
--  LOOP
--    EXECUTE 'SELECT interv FROM temp WHERE NOT visited' INTO r;
--    IF r IS NULL THEN EXIT; END IF;
--    c := r;
--    LOOP
--      EXECUTE 'SELECT interv FROM temp WHERE interv && '
--        || c::text || ' AND NOT visited' INTO u;
--      IF u IS NULL THEN EXIT; END IF;
--      EXECUTE 'UPDATE temp SET visited=TRUE WHERE interv ~= ' || u::text;
--      c := iint_join(c, u);
--    END LOOP;
--    RETURN NEXT c;
--  END LOOP;
--  DROP INDEX temp_idx;
--  DROP TABLE temp;
--  RETURN;
--END;
--$$ LANGUAGE plpgsql IMMUTABLE STRICT;


-- ==========================================================================
-- Utils
-- ==========================================================================

CREATE OR REPLACE FUNCTION revcomp (text)
  RETURNS text
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

-- revcomp(seq, complement?)
CREATE OR REPLACE FUNCTION revcomp (text, bool)
  RETURNS text
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;


-- ==========================================================================
-- Stree
-- ==========================================================================

CREATE TYPE stree;

CREATE OR REPLACE FUNCTION stree_in (cstring)
  RETURNS stree
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION stree_out (stree)
  RETURNS cstring
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION stree_recv (internal)
  RETURNS stree
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION stree_send (stree)
  RETURNS bytea
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE TYPE stree (
  input = stree_in,
  output = stree_out,
  receive = stree_recv,
  send = stree_send,
  internallength = VARIABLE,
  alignment = int4,
  storage = external
);

CREATE OR REPLACE FUNCTION stree_bytea (bytea)
  RETURNS stree
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE CAST (bytea AS stree)
  WITH FUNCTION stree_bytea(bytea) AS IMPLICIT;

CREATE CAST (stree AS bytea)
  WITH FUNCTION stree_send(stree) AS IMPLICIT;

CREATE OR REPLACE FUNCTION stree_build (text)
  RETURNS stree
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE CAST (text AS stree)
  WITH FUNCTION stree_build(text) AS IMPLICIT;

CREATE OR REPLACE FUNCTION stree_text (stree)
  RETURNS text
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE CAST (stree AS text)
  WITH FUNCTION stree_text(stree) AS IMPLICIT;

-- stree_subtext(stree, start, len)
CREATE OR REPLACE FUNCTION stree_subtext (stree, integer, integer)
  RETURNS text
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

CREATE TYPE streematch AS
  (matchlen integer, refstart integer, querystart integer);

-- maxmatch(stree, query, uniqueinref?, minmatchlen):
--  o if uniqueinref is true, maxmatch is analogous to mumcand, otherwise to
--    maxmatch;
--  o if minmatchlen==0, then minmatchlen = length(query);
CREATE OR REPLACE FUNCTION maxmatch (stree, text, bool, integer)
  RETURNS SETOF streematch
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

-- table function analogous to maxmatch: can only be used in queries like
-- "select * from maxmatch0(...)".
CREATE OR REPLACE FUNCTION maxmatch0 (stree, text, bool, integer)
  RETURNS SETOF streematch
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;

-- maxmatchcount(stree, query, uniqueinref?, minmatchlen, anymatch?):
--  o if uniqueinref is true, mum is analogous to mumcand, otherwise to
--    maxmatch;
--  o if minmatchlen==0, then minmatchlen = length(query);
--  o returns number of matches if anymatch == false; o.w., return 1 if there
--    is any match and 0 otherwise.
CREATE OR REPLACE FUNCTION maxmatchcount (stree, text, bool, integer, bool)
  RETURNS integer
  AS '$libdir/postbio'
  LANGUAGE C IMMUTABLE STRICT;



-- ==========================================================================
--
-- Copyright (c) 2008 Luis Carvalho
--
-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation files
-- (the "Software"), to deal in the Software without restriction,
-- including without limitation the rights to use, copy, modify,
-- merge, publish, distribute, sublicense, and/or sell copies of the
-- Software, and to permit persons to whom the Software is furnished
-- to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
-- BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
-- ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
-- CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--
-- ==========================================================================

/* vim: set syn=sql: */
